@@ -4,14 +4,18 @@ OUTFILE="$1"
 VAR16MODE="$2"
 
 # Extract version info
-if [ -d .git -o -f .git ]; then
-    VERSION="`git describe --tags --long --dirty`"
-elif [ -f .version ]; then
-    VERSION="`cat .version`"
+if [ -z "$BUILD_VERSION" ]; then
+    if [ -d .git -o -f .git ]; then
+        VERSION="`git describe --tags --long --dirty`"
+    elif [ -f .version ]; then
+        VERSION="`cat .version`"
+    else
+        VERSION="?"
+    fi
+    VERSION="${VERSION}-`date +"%Y%m%d_%H%M%S"`-`hostname`"
 else
-    VERSION="?"
+    VERSION="$BUILD_VERSION"
 fi
-VERSION="${VERSION}-`date +"%Y%m%d_%H%M%S"`-`hostname`"
 echo "Version: ${VERSION}"
 
 # Build header file
