@@ -26,11 +26,10 @@ struct putcinfo {
  * Debug output
  ****************************************************************/
 
-// Setup debugging port(s).
 void
-debug_preinit(void)
+debug_banner(void)
 {
-    serial_debug_preinit();
+    dprintf(1, "SeaBIOS (version %s)\n", VERSION);
 }
 
 // Write a character to debug port(s).
@@ -77,6 +76,7 @@ screenc(char c)
     br.flags = F_IF;
     br.ah = 0x0e;
     br.al = c;
+    br.bl = 0x07;
     call16_int(0x10, &br);
 }
 
